
#include "unix.h"

typedef struct _Point {
    float x, y, z;
} Point;

typedef struct {
    float x, y, z;
} PEXVector;

typedef float PEXMatrix[4][4];

#define INPUT
#define OUTPUT

/* constants for PEXRotate */
#define PEXXAxis                1
#define PEXYAxis                2
#define PEXZAxis                3
#define PEXBadAxis              10

#define DEGRAD( _a )  ( (float)( _a ) * 3.1415926 / 180.0 )
#define RADDEG( _a )  ( (float)( _a ) * 180.0 / 3.1415926 )

int  PEXRotate();
void PEXTranslate();
void PEXIdentityMatrix();
void PEXMatrixMult();
void PEXVectorMult();

Cone( point1, radius1, point2, radius2, depth,
      clockwise, top, bottom, primitive )
Point point1;
float radius1;
Point point2;
float radius2;
int   depth;
char clockwise;
char top, bottom;
int primitive;
{
    PEXMatrix T, Ry, Rx, m;
    PEXMatrix mT, mRy, mRx;
    float P1[4], P1p[4], P1pp[4], P2[4], P2p[4], P2pp[4], P2ppp[4];
    float temp[4];
    PEXVector t;
    double cosTheta, cosPhi, D;
    double angle_theta, angle_phi;

    int i, j;        /* i/length    j/theta */
    float is, js;
    float theta, step, lstep;
    float flengthSize, fthetaSize;
    int lengthSize, thetaSize;
    float yaxis, rstep;
    Point cone[130][260];
    FILE *fp;
    extern char *filename;

    t.x = -point1.x;    t.y = -point1.y;    t.z = -point1.z;
    PEXTranslate( &t, T );
    t.x =  point1.x;    t.y =  point1.y;    t.z =  point1.z;
    PEXTranslate( &t, mT );

    P1[0] = point1.x;   P1[1] = point1.y;   P1[2] = point1.z;  P1[3] = 1.;
    PEXVectorMult( T, P1, P1p );

    P2[0] = point2.x;   P2[1] = point2.y;   P2[2] = point2.z;  P2[3] = 1.;
    PEXVectorMult( T, P2, P2p );

    D = sqrt( (point2.z - point1.z) * (point2.z - point1.z) +
              (point2.x - point1.x) * (point2.x - point1.x) );
    cosTheta = -( point2.z - point1.z ) / D;
    if( point2.x > point1.x ) angle_theta = acos( cosTheta );
    if( point2.x < point1.x ) angle_theta = -acos( cosTheta );
    PEXRotate( PEXYAxis, angle_theta, Ry );
    PEXRotate( PEXYAxis, -angle_theta, mRy );
    PEXVectorMult( Ry, P2p, P2pp );

    D = sqrt( (point2.z - point1.z) * (point2.z - point1.z) +
              (point2.y - point1.y) * (point2.y - point1.y) +
              (point2.x - point1.x) * (point2.x - point1.x) );
    cosPhi = -P2pp[2] / D;
    if( point2.y > point1.y ) angle_phi = M_PI / 2. + (-acos( cosPhi ));
    if( point2.y < point1.y ) angle_phi = M_PI / 2. + acos( cosPhi );
    PEXRotate( PEXXAxis, angle_phi, Rx );
    PEXRotate( PEXXAxis, -angle_phi, mRx );
    PEXVectorMult( Rx, P2pp, P2 );

    yaxis = P2[1];
    step = (float)(M_PI / pow( (double)2, (double)depth ));
    lstep = (float)(yaxis / pow( (double)2, (double)depth ));
    flengthSize = (yaxis / lstep) + 1.;
    fthetaSize = (2 * M_PI / step) + 1.;
    lengthSize = (int)flengthSize;
    thetaSize = (int)fthetaSize;

    rstep = fabs( radius1 - radius2 ) / ( lengthSize - 1 );
    for( i = 0; i < lengthSize; i++ ) {
        for( j = 0; j < thetaSize; j++ ) {
            is = i * lstep;   js = j * step;
            cone[i][j].x = (radius1 - ( i * rstep )) * cos( js ) +
                ( P1[0] + i * (fabs(P2[0] - P1[0]) / ( lengthSize - 1 )) );
            cone[i][j].y = is;
            cone[i][j].z = (radius1 - ( i * rstep )) * sin( js ) +
                ( P1[2] + i * (fabs(P2[2] - P1[2]) / ( lengthSize - 1 )) );
            temp[0] = cone[i][j].x;
            temp[1] = cone[i][j].y;
            temp[2] = cone[i][j].z;
            temp[3] = 1.;
            PEXVectorMult( mRx, temp, temp );
            PEXVectorMult( mRy, temp, temp );
            PEXVectorMult(  mT, temp, temp );
            cone[i][j].x = temp[0];
            cone[i][j].y = temp[1];
            cone[i][j].z = temp[2];
        }
    }

    fp = fopen( filename, "w" );

    fprintf( fp, "%d\t", lengthSize * thetaSize );
    if( top && bottom )
        fprintf( fp, "%d\t", (lengthSize-1) * (thetaSize-1) + 2 );
    else if( top || bottom )
        fprintf( fp, "%d\t", (lengthSize-1) * (thetaSize-1) + 1 );
    else
        fprintf( fp, "%d\t", (lengthSize-1) * (thetaSize-1) );
    if( top && bottom )
        fprintf( fp, "%d\n", 4 * (lengthSize-1) * (thetaSize-1) + 2 * thetaSize );
    else if( top || bottom )
        fprintf( fp, "%d\n", 4 * (lengthSize-1) * (thetaSize-1) + thetaSize );
    else
        fprintf( fp, "%d\n", 4 * (lengthSize-1) * (thetaSize-1) );

    /* Print the Coordinates of the points */
    for( i = 0; i < lengthSize; i++ )
        for( j = 0; j < thetaSize; j++ )
            fprintf( fp, "%f %f %f\n", cone[i][j].x,cone[i][j].y,cone[i][j].z );

    /* Print the Conectivity data of the points */
    if( clockwise )  {
        if( bottom ) {
            fprintf( fp, "%d ", thetaSize );
            for( i = thetaSize-1; i >= 0; i-- ) {
                fprintf( fp, "%d ", 0*thetaSize+i+1 );
            }
            fprintf( fp, "\n" );
        }
        for( i = 0; i < thetaSize-1; i++ ) {
            for( j = 0; j < lengthSize-1; j++ ) {
                fprintf( fp, "4 %d %d %d %d\n",
                           j*thetaSize+i+1,     j*thetaSize+(i+1)+1,
                   (j+1)*thetaSize+(i+1)+1, (j+1)*thetaSize+i+1 );
            }
        }
        if( top ) {
            fprintf( fp, "%d ", thetaSize );
            for( i = 0; i < thetaSize; i++ ) {
                fprintf( fp, "%d ", (lengthSize-1)*thetaSize+i+1 );
            }
            fprintf( fp, "\n" );
        }
    }
    else {    /* counterClockwise */
        if( bottom ) {
            fprintf( fp, "%d ", thetaSize );
            for( i = 0; i < thetaSize; i++ ) {
                fprintf( fp, "%d ", 0*thetaSize+i+1 );
            }
            fprintf( fp, "\n" );
        }
        for( i = 0; i < thetaSize-1; i++ ) {
            for( j = 0; j < lengthSize-1; j++ ) {
                fprintf( fp, "4 %d %d %d %d\n",
                           j*thetaSize+i+1, (j+1)*thetaSize+i+1,
                   (j+1)*thetaSize+(i+1)+1, j*thetaSize+(i+1)+1 );
            }
        }
        if( top ) {
            fprintf( fp, "%d ", thetaSize );
            for( i = thetaSize-1; i >= 0; i-- ) {
                fprintf( fp, "%d ", (lengthSize-1)*thetaSize+i+1 );
            }
            fprintf( fp, "\n" );
        }
    }
    fclose( fp );
}


int PEXRotate (axis, angle, matrix_return)

INPUT int        axis;
INPUT double        angle;
OUTPUT PEXMatrix    matrix_return;

{
    double    sine;
    double    cosine;

    sine = sin (angle);
    cosine = cos (angle);

    switch (axis)
    {
    case PEXXAxis:
        matrix_return[0][0] = 1.0;
        matrix_return[0][1] = 0.0;
        matrix_return[0][2] = 0.0;
        matrix_return[0][3] = 0.0;

        matrix_return[1][0] = 0.0;
        matrix_return[1][1] = cosine;
        matrix_return[1][2] = -sine;
        matrix_return[1][3] = 0.0;

        matrix_return[2][0] = 0.0;
        matrix_return[2][1] = sine;
        matrix_return[2][2] = cosine;
        matrix_return[2][3] = 0.0;

        matrix_return[3][0] = 0.0;
        matrix_return[3][1] = 0.0;
        matrix_return[3][2] = 0.0;
        matrix_return[3][3] = 1.0;
        break;

    case PEXYAxis:
        matrix_return[0][0] = cosine;
        matrix_return[0][1] = 0.0;
        matrix_return[0][2] = sine;
        matrix_return[0][3] = 0.0;

        matrix_return[1][0] = 0.0;
        matrix_return[1][1] = 1.0;
        matrix_return[1][2] = 0.0;
        matrix_return[1][3] = 0.0;

        matrix_return[2][0] = -sine;
        matrix_return[2][1] = 0.0;
        matrix_return[2][2] = cosine;
        matrix_return[2][3] = 0.0;

        matrix_return[3][0] = 0.0;
        matrix_return[3][1] = 0.0;
        matrix_return[3][2] = 0.0;
        matrix_return[3][3] = 1.0;
        break;

    case PEXZAxis:
        matrix_return[0][0] = cosine;
        matrix_return[0][1] = -sine;
        matrix_return[0][2] = 0.0;
        matrix_return[0][3] = 0.0;

        matrix_return[1][0] = sine;
        matrix_return[1][1] = cosine;
        matrix_return[1][2] = 0.0;
        matrix_return[1][3] = 0.0;

        matrix_return[2][0] = 0.0;
        matrix_return[2][1] = 0.0;
        matrix_return[2][2] = 1.0;
        matrix_return[2][3] = 0.0;

        matrix_return[3][0] = 0.0;
        matrix_return[3][1] = 0.0;
        matrix_return[3][2] = 0.0;
        matrix_return[3][3] = 1.0;
        break;

    default:
        return (PEXBadAxis);     /* error - invalid axis specifier */
    }

    return (0);
}


void PEXTranslate (trans_vector, matrix_return)

INPUT PEXVector        *trans_vector;
OUTPUT PEXMatrix    matrix_return;

{
    matrix_return[0][0] = 1.0;
    matrix_return[0][1] = 0.0;
    matrix_return[0][2] = 0.0;
    matrix_return[0][3] = trans_vector->x;

    matrix_return[1][0] = 0.0;
    matrix_return[1][1] = 1.0;
    matrix_return[1][2] = 0.0;
    matrix_return[1][3] = trans_vector->y;

    matrix_return[2][0] = 0.0;
    matrix_return[2][1] = 0.0;
    matrix_return[2][2] = 1.0;
    matrix_return[2][3] = trans_vector->z;

    matrix_return[3][0] = 0.0;
    matrix_return[3][1] = 0.0;
    matrix_return[3][2] = 0.0;
    matrix_return[3][3] = 1.0;
}


void PEXIdentityMatrix (matrix_return)

OUTPUT PEXMatrix    matrix_return;

{
    matrix_return[0][0] = 1.0;
    matrix_return[0][1] = 0.0;
    matrix_return[0][2] = 0.0;
    matrix_return[0][3] = 0.0;

    matrix_return[1][0] = 0.0;
    matrix_return[1][1] = 1.0;
    matrix_return[1][2] = 0.0;
    matrix_return[1][3] = 0.0;

    matrix_return[2][0] = 0.0;
    matrix_return[2][1] = 0.0;
    matrix_return[2][2] = 1.0;
    matrix_return[2][3] = 0.0;

    matrix_return[3][0] = 0.0;
    matrix_return[3][1] = 0.0;
    matrix_return[3][2] = 0.0;
    matrix_return[3][3] = 1.0;
}


void PEXMatrixMult (matrix1, matrix2, matrix_return)

INPUT PEXMatrix     matrix1;
INPUT PEXMatrix     matrix2;
OUTPUT PEXMatrix    matrix_return;

{
    register float  *r;
    register int    i;

    if ((matrix_return != matrix1) && (matrix_return != matrix2))
    {
        for (i = 0; i < 4; i++)
        {
            r = matrix1[i];
            matrix_return[i][0] = r[0] * matrix2[0][0] + r[1] * matrix2[1][0] +
                r[2] * matrix2[2][0] + r[3] * matrix2[3][0];
            matrix_return[i][1] = r[0] * matrix2[0][1] + r[1] * matrix2[1][1] +
                r[2] * matrix2[2][1] + r[3] * matrix2[3][1];
            matrix_return[i][2] = r[0] * matrix2[0][2] + r[1] * matrix2[1][2] +
                r[2] * matrix2[2][2] + r[3] * matrix2[3][2];
            matrix_return[i][3] = r[0] * matrix2[0][3] + r[1] * matrix2[1][3] +
                r[2] * matrix2[2][3] + r[3] * matrix2[3][3];
        }
    }
    else
    {
        register float  *src, *dst;
        PEXMatrix   temp;

        for (i = 0; i < 4; i++)
        {
            r = matrix1[i];
            temp[i][0] = r[0] * matrix2[0][0] + r[1] * matrix2[1][0] +
                r[2] * matrix2[2][0] + r[3] * matrix2[3][0];
            temp[i][1] = r[0] * matrix2[0][1] + r[1] * matrix2[1][1] +
                r[2] * matrix2[2][1] + r[3] * matrix2[3][1];
            temp[i][2] = r[0] * matrix2[0][2] + r[1] * matrix2[1][2] +
                r[2] * matrix2[2][2] + r[3] * matrix2[3][2];
            temp[i][3] = r[0] * matrix2[0][3] + r[1] * matrix2[1][3] +
                r[2] * matrix2[2][3] + r[3] * matrix2[3][3];
        }

        src = (float *) temp;
        dst = (float *) matrix_return;
        for (i = 0; i < 16; i++)
            *dst++ = *src++;
    }
}


void PEXVectorMult (matrix2, matrix1, matrix_return)

INPUT PEXMatrix matrix2;
INPUT float     matrix1[4];
OUTPUT float    matrix_return[4];

{
    register int i, j;

    if (matrix_return != matrix1)
    {
        for ( i=0; i<4; i++ )
        {
            matrix_return[i] = 0.;
            for ( j=0; j<4; j++ )
               matrix_return[i] = matrix_return[i] + matrix2[i][j] * matrix1[j];
        }
    }
    else
    {
        register float  *src, *dst;
        float temp[4];

        for ( i=0; i<4; i++ )
        {
            temp[i] = 0.;
            for ( j=0; j<4; j++ )
                temp[i] = temp[i] + matrix2[i][j] * matrix1[j];
        }

        src = (float *) temp;
        dst = (float *) matrix_return;
        for (i = 0; i < 4; i++)
            *dst++ = *src++;
    }
}
