
#include <stdio.h>
#include <math.h>
#include "off.h"

/* Vertex Order information */
#define CLOCKWISE			1
#define COUNTERCLOCKWISE	2

static char *ProgramName;
static int toascii = 1, addnormals = 0;
char *infilename = NULL, *outfilename;

typedef struct _Vertex {
    float x, y, z;
} Vertex;

typedef Vertex Vector;

typedef struct _Edge {
    int e;
} Edge;

typedef struct _Facet {
    int count;
    Edge *edges;
} Facet;

Usage() { fprintf(stderr, "usage: %s inobj \n", ProgramName ); }


ParseCommandLine(argc, argv)
int argc;
char *argv[];
{
    int ArgsParsed = 0;
    char *sw;

    ProgramName = argv[ArgsParsed++];

    while ( ArgsParsed < argc ) {
        if (argv[ArgsParsed][0] == '-')
        {
            sw = argv[ArgsParsed++] + 1;
            for( ; *sw != '\0'; sw++) {
                switch (*sw)
                {
                default:
                    fprintf(stderr, "Unknown option '%s'\n\n", *sw );
                    Usage();
                    exit(-1);
                }
            }
        } else {
            if (infilename != NULL) {
/*              if (outfilename != NULL) { */
                    Usage(); 
                    exit(-1);
/*              } */
/*              outfilename = argv[ArgsParsed++]; */
            }  else
                infilename = argv[ArgsParsed++];
        }
    }
    if (infilename == NULL /* || outfilename == NULL */ ) {
        Usage(); 
        exit(-1);
    }
}


basename(fullname, dir, base, ext)
char	*fullname;
char	*dir;
char	*base;
char	*ext;
{
    int		i, j;

    i = strlen(fullname);
    while ((i >= 0) && (fullname[i] != '.')) i--;
    if (i < 0)
    { 
        i = 0; 
        ext[0] = '\0'; 
    }
    else
        strcpy(ext, &(fullname[i + 1]));

    j = i;
    while ((j >= 0) && (fullname[j] != '/')) j--;
    if (j < 0)
    { 
        dir[0] = '\0'; 
    }
    else
    {
        strncpy(dir, fullname, j);
        dir[j] = '\0';
    }

    if ((j <= 0) && (i == 0))
        strcpy(base, fullname);
    else
    {
        strncpy(base, &(fullname[j + 1]), i - j - 1);
        base[i - j - 1] = '\0';
    }
}


/* Some types needed for the normal generation.  We should probably
**  stick these in a header file.
*/
#if( 0 ) 
typedef float float32_t;
typedef short int16_t;
typedef long  int32_t;
#endif		/*  */
typedef struct { 
    float x, y, z; 
} Point;
typedef struct { 
    float dx, dy, dz; 
} Normal;
typedef struct { 
    long count; 
    Normal n[1]; 
} NormalProperty;


AddNormals( basename, obj )
char *basename;
OFFObjDesc *obj;
{
    OFFProperty *gprop = NULL, *pnprop = NULL, *op;
    int nv, np, ne;
    NormalProperty *normals;
    Normal *normal;
    Point *vertices;
    short *edges;
    short *nedges;
    int clockwise = 1;
    int i, start, end, j, vi, vj, p, n;
    char *ptr;
    double len;

    /*  Find the "geometery" property and the vertex_order property*/
    for( op = obj->FirstProp; op; op = op->NextProp) {
        if (strcmp( op->PropName, "geometry") == 0)
            gprop = op;
        else if (strcmp( op->PropName, "vertex_order") == 0) {
            if (strcmp( *((char **) op->PropData), "clockwise"))
                clockwise = 1;
            else if (strcmp( *((char **) op->PropData), "counter-clockwise"))
                clockwise = 0;
            else {
                fprintf( stderr, "Invalid vertex order: %s\n", *((char **) op->PropData));
                exit(-1);
            }
        }
    }

    /* Remove any existing polygon normals property */
    OFFRemoveProperty( obj, "polygon_normals" );

    /* Get the geometry data */
    ptr = gprop->PropData;

    /* extract the counts */
    nv = *( (long *) ptr); 
    ptr += sizeof(long);
    np = *( (long *) ptr); 
    ptr += sizeof(long);
    ne = *( (long *) ptr); 
    ptr += sizeof(long);


    /* Get the vertex data */
    vertices = (Point *) ptr;
    ptr += 3*nv*sizeof(float);

    /* Get the edge-per-polygon counts */
    nedges = (short *) ptr;
    ptr += np * sizeof(short);

    /* Get the edge data */
    edges = (short *) ptr;
    ptr += ne *sizeof(short);

    /* Allocate memory for the normals */
    normals = (NormalProperty *) malloc( np * sizeof(Normal) + sizeof(short) );
    if (normals == NULL) {
        perror( "Out of memory!\n" );
        exit(-1);
    }
    normals->count = np;

    /* Add a new polygon_normals property */
    pnprop = OFFAddProperty( obj );
    if (pnprop == NULL)  {
        perror( "Out of memory!\n" );
        exit(-1);
    }

    /* Set up the property */
    strcpy( pnprop->PropName, "polygon_normals" );
    pnprop->PropType = OFF_GENERIC_DATA;
    strcpy( pnprop->PropFileName, basename );
    strcat( pnprop->PropFileName, ".pnorm" );
    pnprop->PropCount = np;
    strcpy( pnprop->DataFormat, "fff" );
    pnprop->PropData = (char *) normals;

    /* Loop through all the polygons computing normals as we go */
    i=0;
    for(p=0; p<np; p++) {
        n = nedges[p];
        start = i;
        end = i + n;
        normal = normals->n +p;
        normal->dx = normal->dy = normal->dz = 0.0;
        for( j=i+1; i < end; i++, j++) {
            if (j == end) j = start;
            vi = edges[ i ] -1;
            vj = edges[ j ] -1;

            /* Martin Newell's method:  Newman & Sproull page 499
      ** More robust than the 3 point method.  
      */
            normal->dx += (vertices[vi].y - vertices[vj].y)*
                (vertices[vi].z + vertices[vj].z);
            normal->dy += (vertices[vi].z - vertices[vj].z)*
                (vertices[vi].x + vertices[vj].x);
            normal->dz += (vertices[vi].x - vertices[vj].x)*
                (vertices[vi].y + vertices[vj].y);
        }
        len = sqrt( normal->dx * normal->dx +
            normal->dy * normal->dy +
            normal->dz * normal->dz );
        if (len > 0) {		/* Normalize length */
            normal->dx /= len;
            normal->dy /= len;
            normal->dz /= len;
        }
        if (!clockwise) {		/* Flip the normal */
            normal->dx = -normal->dx;
            normal->dy = -normal->dy;
            normal->dz = -normal->dz;
        }
    }
}


main(argc, argv)
int		argc;
char	*argv[];
{
    OFFObjDesc	obj;
    char	inbase[80], outbase[80];
    char	indir[80], outdir[80];
    char	inext[80], outext[80];
    char	newname[80];
    int		i, j;
    int		type;
    char	dir[80], base[80], ext[80];
    OFFProperty	*pProp;

    long nver, nfac, nedg;
    unsigned int nvertices, nfacets, nedges;
    Vertex *vertices;
    Facet *facets;
    int vertex_order = COUNTERCLOCKWISE;

    ParseCommandLine( argc, argv );

    if (-1 == OFFReadObj(&obj, infilename ) )
        exit(-1);

    basename(infilename, indir, inbase, inext);
    basename(outfilename, outdir, outbase, outext);
    if (addnormals)
        AddNormals( outbase, &obj );

    pProp = obj.FirstProp;
    while (pProp != NULL)
    {
        if (pProp->PropFileName != NULL)
        {
            basename(pProp->PropFileName, dir, base, ext);
            if (toascii)
            {
                strcpy(newname, outbase);
                strcat(newname, ".");

                /*  The following kludge will not work if there are valid  */
                /*  ASCII file extensions that begin with 'b'.  Right now  */
                /*  we assume that if the extension begins with 'b', the   */
                /*  file is in binary.                                     */

                if (ext[0] == 'b')
                    strcat(newname, &(ext[1]));
                else
                    strcat(newname, ext);
            }
            else
            {
                strcpy(newname, outbase);
                strcat(newname, ".b");
                strcat(newname, ext);
            }
            strcpy(pProp->PropFileName, newname);
        }
        pProp = pProp->NextProp;
    }

    if (toascii)
        type = OFF_ASCII;
    else
        type = OFF_BINARY;

    pProp = obj.FirstProp;
    while(pProp != NULL)
    {
        char  *ptr;    long  *lptr;    float *fptr;    short *sptr;

        /* WriteProperty(ObjFile, pProp, Dir, FileType); */
        if( strcmp( pProp->PropName, "geometry" ) == 0 ) {
            if( pProp->PropType == OFF_INDEXED_POLY_DATA ) {
                ptr  = pProp->PropData;
                lptr = (long *) ptr;
                nver = *lptr++;
                nfac = *lptr++;
                nedg = *lptr++;
                ptr  = (char *) lptr;

                nvertices = nver;
                nfacets   = nfac;
                nedges    = nedg;

                if( (vertices = (Vertex *)malloc( nvertices * sizeof(Vertex) ))
                 == (Vertex *)0 ) {
                    fprintf( stderr, "Cant allocate Vertex 2\n" );
                    exit(1);
                }
                if( (facets = (Facet *)malloc( nfacets * sizeof(Facet) )) ==
                    (Facet *)0 ) {
                    fprintf( stderr, "Cant allocate Facet\n" );
                    exit(1);
                }

                fptr = (float *) ptr;
                for(i = 0; i < nvertices; i++) {
                    vertices[i].x = *fptr++;
                    vertices[i].y = *fptr++;
                    vertices[i].z = *fptr++;
                }
                ptr = (char *) fptr;

                sptr = (short *) ptr;
                for(j = 0; j < nfacets; j++) {
                    facets[j].count = *sptr++;
                    if( (facets[j].edges = (Edge *)malloc( facets[j].count *
                                                           sizeof(Edge) )) ==
                        (Edge *)0 ) {
                        fprintf( stderr, "Cant allocate Edge 2\n" );
                        exit(1);
                    }
                }
                for(j = 0; j < nfacets; j++)
                    for(i = 0; i < facets[j].count; i++)
                        facets[j].edges[i].e = *sptr++;
                ptr = (char *) sptr;
            }
        }
        pProp = pProp->NextProp;
    }

    printf( "#declare te = pigment { color red 1 green 0 blue 0 }\n" );

    printf( "camera\n{\n location  <5.0,4.5,4.5>\n" );
    printf( "look_at <0, 0, 0>\n}\n" );

    printf( "object\n{\n light_source {\n" );
    printf( "  <250, 500, 500> color red 1 green 1 blue 1\n }\n" );
    printf( "}\n" );
    printf( "object\n{\n light_source {\n" );
    printf( "  <-250, 500, 500> color red 1 green 1 blue 1\n }\n" );
    printf( "}\n" );

    for( j=0; j<nfacets; j++ ) 
    {
        Vertex *polygon = (Vertex *)malloc( facets[j].count * sizeof(Vertex) );
        /* printf( "p %d\n", facets[j].count ); */
        for( i=0; i<facets[j].count; i++ )
        {
/*
            printf( "%.6f %.6f %.6f\n",
                vertices[facets[j].edges[i].e - 1].x,
                vertices[facets[j].edges[i].e - 1].y,
                vertices[facets[j].edges[i].e - 1].z );
*/
            polygon[i].x = vertices[facets[j].edges[i].e - 1].x;
            polygon[i].y = vertices[facets[j].edges[i].e - 1].y;
            polygon[i].z = vertices[facets[j].edges[i].e - 1].z;
        }
        draw_poly( facets[j].count, polygon, 2 /* POV */ );
        free( polygon );
        polygon = (Vertex *)NULL;
    }
    /* OFFWriteObj(&obj, outfilename, outdir, type); */
}
